import enum

__all__ = [
    "Options"
]

class Options(enum.Enum):
    BackgroundAttachment = "background-attachment"
    BackgroundBlendMode = "background-blend-mode"
    BackgroundColor = "background-color"
    BackgroundImage = "background-image"
    BackgroundPosition = "background-position"
    BackgroundRepeat = "background-repeat"
    Border = "border"
    BorderBottom = "border-bottom"
    BorderBottomColor = "border-bottom-color"
    BorderBottomStyle = "border-bottom-style"
    BorderBottomWidth = "border-bottom-width"
    BorderCollapse = "border-collapse"
    BorderColor = "border-color"
    BorderLeft = "border-left"
    BorderLeftColor = "border-left-color"
    BorderLeftStyle = "border-left-style"
    BorderLeftWidth = "border-left-width"
    BorderRight = "border-right"
    BorderRightColor = "border-right-color"
    BorderRightStyle = "border-right-style"
    BorderRightWidth = "border-right-width"
    BorderSpacing = "border-spacing"
    BorderStyle = "border-style"
    BorderTop = "border-top"
    BorderTopColor = "border-top-color"
    BorderTopStyle = "border-top-style"
    BorderTopWidth = "border-top-width"
    BorderWidth = "border-width"
    Bottom = "bottom"
    BreakAfter = "break-after"
    BreakBefore = "break-before"
    BreakInside = "break-inside"
    CaptionSide = "caption-side"
    Clear = "clear"
    Clip = "clip"
    Color = "color"
    Content = "content"
    CounterIncrement = "counter-increment"
    CounterReset = "counter-reset"
    Cursor = "cursor"
    Direction = "direction"
    Display = "display"
    EmptyCells = "empty-cells"
    Float = "float"
    Font = "font"
    FontFamily = "font-family"
    FontSize = "font-size"
    FontStyle = "font-style"
    FontVariant = "font-variant"
    FontWeight = "font-weight"
    Grid = "grid"
    GridArea = "grid-area"
    GridAutoColumns = "grid-auto-columns"
    GridAutoFlow = "grid-auto-flow"
    GridAutoRows = "grid-auto-rows"
    GridColumn = "grid-column"
    GridColumnEnd = "grid-column-end"
    GridColumnGap = "grid-column-gap"
    GridColumnStart = "grid-column-start"
    GridGap = "grid-gap"
    GridRow = "grid-row"
    GridRowEnd = "grid-row-end"
    GridRowGap = "grid-row-gap"
    GridRowStart = "grid-row-start"
    GridTemplate = "grid-template"
    GridTemplateAreas = "grid-template-areas"
    GridTemplateColumns = "grid-template-columns"
    GridTemplateRows = "grid-template-rows"
    Height = "height"
    Left = "left"
    LetterSpacing = "letter-spacing"
    LineHeight = "line-height"
    ListStyle = "list-style"
    ListStyleImage = "list-style-image"
    ListStylePosition = "list-style-position"
    ListStyleType = "list-style-type"
    Margin = "margin"
    MarginBottom = "margin-bottom"
    MarginLeft = "margin-left"
    MarginRight = "margin-right"
    MarginTop = "margin-top"
    MaxHeight = "max-height"
    MaxWidth = "max-width"
    MinHeight = "min-height"
    MinWidth = "min-width"
    MixBlendMode = "mix-blend-mode"
    Outline = "outline"
    OutlineColor = "outline-color"
    OutlineStyle = "outline-style"
    OutlineWidth = "outline-width"
    Overflow = "overflow"
    Padding = "padding"
    PaddingBottom = "padding-bottom"
    PaddingLeft = "padding-left"
    PaddingRight = "padding-right"
    PaddingTop = "padding-top"
    PageBreakAfter = "page-break-after"
    PageBreakBefore = "page-break-before"
    PageBreakInside = "page-break-inside"
    Position = "position"
    Quotes = "quotes"
    Right = "right"
    ScrollBehavior = "scroll-behavior"
    TableLayout = "table-layout"
    TextAlign = "text-align"
    TextDecoration = "text-decoration"
    TextDecorationThickness = "text-decoration-thickness"
    TextIndent = "text-indent"
    TextTransform = "text-transform"
    Top = "top"
    UnicodeBidi = "unicode-bidi"
    VerticalAlign = "vertical-align"
    Visibility = "visibility"
    WhiteSpace = "white-space"
    Width = "width"
    WordSpacing = "word-spacing"
    ZIndex = "z-index"
